/* GETGUI (II) .........  API */

/* Note: "sethandle", "listloc", "object id" all basically refer to the same thing.. */

/* Multiple instances of any of these objects is supported by the API, even though getgui(1) 
 * doesn't really do this.  Taborder between multiple textfields is supported.
 */

#include "plg.h"

#define MAXPICK 400
#define MAXVALLEN 128
#define MAXSELECTROWS 60
#define MAXSELECTROWLEN 128

#define BUTTON 'B'
#define TEXTFIELD 'T'
#define NAMESELECT 'N'
#define PULLDOWN 'P'		/* the "closed" pulldown bar */
#define PULLDOWN_DEPLOYED 'D'	/* the rectangle that the deployed pulldown list occupies (non interactive) */
#define PULLDOWN_CHOICE 'C'	/* an individual pulldown choice */
#define MISC 'X'		/* general zone, non-interactive */

#define FLASH_uS 100000

#define INACTIVE 0
#define ACTIVE 1

/* pickreps */
#define ENTIRE_TEXT 0   /* return the entire text */
#define FIRST_TOKEN 1   /* return only the first whitespace-delimited token - available w/ pulldowns & selectboxes */
#define INDEX_NUM 2     /* return the position # of the selection, first=1 - available w/ buttons & selectboxes */

#define BUTTON_COLOR "yellow"	/* default button color */

/* keys */
#define LEFT 551        /* left arrow */
#define UP 552          /* up arrow */
#define RIGHT 553       /* right arrow */
#define DOWN 554        /* down arrow */
#define HOME 550        /* HOME */
#define CTRL_A 1        /* same as HOME */
#define END  557        /* END */
#define CTRL_E 5        /* same as END */
#define CTRL_K 11       /* delete to end of line */
#define CTRL_U 21       /* delete entire line */
#define DEL '\177'      /* DEL */
#define CTRL_D  4       /* ^D */
#define BACKSPACE 8     /* ^H */
#define TAB 9
#define ENTER 13
/* the following may be set to 1 to make DEL=BACKSPACE, 0 if not */
#define DEL_EQUIV_BACKSPACE 0


/* ------- */
/* general */
/* ------- */

char tmpfilename[256];		/* temp file name.. draw commands written to this file */

/* window location in pixels */
int win_x = -1, win_y = -1;    /* -1 = centered on screen */

/* fixed window size in inches */
double winfixw = -1.0, winfixh = -1.0;

/* window margin amount, in inches */
double winxmarg = 0.2, winymarg = 0.2;

/* list separator character, for multi-choice responses.. */
char listsep[5] = ",";

/* filename for a file that be used to clear an asynchronous 'in progress' message.. */
char *clrfile = NULL;

/* async - if 1 output a line to stdout every 1/2 second.. useful for apps (eg. logo-lx) 
   that can't completely block waiting for getgui response */
int async = 0;

/* current textentry field */
int curtextfield = 0;
/* current cursor location for blink.. */
double cursorx, cursory, cursortopy;
int cursorstate = 0;

/* currently active pulldown */
int pulldown_deployed;

/* translation vector */
int autocrop;  /* 1 = do autocrop and auto center on screen   0 = don't */
double transx, transy;

/* relalign - alignment relative to an existing object */
int relalign = 0;
double x_override, y_override;


/* --------- */
/* pick list */
/* --------- */

/* Pick list BUTTON entries (one entry per button in the button set)
 *	objval = text of the selected button
 *	iprop = 1 if button selected, 0 if not
 *      iprop2 = array location of first button in group
 *      iprop3 = number of buttons in set (first button entry only)
 *	iprop4 = point size of button labels (also affects overall button size)
 *	cprop  = button color
 * Also an additional MISC entry is created for the entire region, for relalign purposes.
 *
 * Pick list TEXTFIELD entries (one pick list entry per textfield)
 *	objval = the keyed input string
 *	iprop = cursor current column position relative to input string 
 *	iprop2 = length (in chars) of visible text field
 *	iprop3 = max allowable length (may be 0 for MAXVALLEN)
 *	iprop4 = 1 if active, 0 if not
 *	cprop  = "P" = password mode
 *
 * Pick list NAMESELECT entries (one pick list entry per selectable name)
 * (pulldown choices also use this scheme)
 *	objval = selectable name
 *	iprop  = 1 if name selected, 0 if not
 *      iprop2 = array location of the first name in group
 *	iprop3 = number of names in group (first entry only)
 * Also an additional MISC entry is created for the entire region, for relalign purposes.
 *
 * Pick list PULLDOWN entries (one pick list entry for the pulldown bar)
 * (there will also be one nameselect entry (see above) for each choice)
 *	objval = selectable name
 *      iprop  = 1 if active, 0 if inactive
 *      iprop3  = number of names in group
 * Also a PULLDOWN_DEPLOYED entry is created for the entire deployed region.
 * So, a pulldown object has these entries:  1: pulldown  2: pulldown_deploy  3-?: nameselect 
 * 
 */

struct picklist {
	double x1, y1, x2, y2; 		/* pickable region */
	char objtype;			/* B = button   T = textfield   N = nameselect box   P,D,C = pulldown */
	char objval[MAXVALLEN+1];	/* contents / label */
	int iprop;			/* integer property 1 - varies depending on objtype */
	int iprop2;			/*  ''      ''      2 - '' */
	int iprop3;			/*  ''      ''      3 - '' */
	int iprop4;			/*  ''      ''      4 - '' */
	char cprop[40];			/*  char property     - varies depending on objtype */
	};

struct picklist pick[MAXPICK];
int npick = 0;


/* ---------- */
/* properties */
/* ---------- */

/* overall */
char bgcolor[40] = "gray(0.9)";

/* properties of all button sets */
double bsep = 0.1;		/* amount of add'l separation between buttons */
double bdressup = 1;		/* dressup to apply to buttons (0=outline, 1=bevel, 2=dropshadow) */
double bbevelsize = 0.2;	/* controls width of beveling*/
char bflashcolor[40] = "gray(0.9)"; /* color of selected buttons */
int btextsize = 12;		/* text size for labels */
char *bcolorlist = "xrgb(ecffec)|xrgb(ffffec)|xrgb(ffecff)|xrgb(ecffff)|xrgb(ffe0d0)|xrgb(c0c0c0)"; /* a canned color list */
				/* note: button color can be specified live for each set (default = yellow) */

/* properties of all text fields */
double tmarg = 0.08;		/* margin above, below, and on either side of entered text */
char tcolor[40] = "white";	/* background color of text field. */
int ttextsize = 12;		/* text size */

/* properties of all name selects (and pulldown choices) */
double nmarg = 0.08;		/* margin above, below, and on either side of box */
char ncolor[40] = "white"; 	/* background color within select box */
int ntextsize = 10; 		/* text size */
char nflashcolor[40] = "darkblue";  /* color for selected items */

/* properties of pulldowns */
char pcolor[40] = "white";	/* background color */
int ptextsize = 10;		/* text size */


/* ---- */
/* misc */
/* ---- */
/* x11 char widths */
double xcharwidths[21] = { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0667014, 0.0667014, 0.0800417, 0.0800417, 0.0933819, 0.0933819,
	0.120062, 0.120062, 0.120062, 0.120062, 0.146743, 0.146743, 0.146743, 0.146743, 0.200104, 0.200104 };
double charw;

char *fullobjmsg = "too many pickable objects\n";
char *badlistlocmsg = "%s: invalid object handle\n";


/* ---------------- */
/* static functions */
/* ---------------- */
static int render(), whichpick(), flashbutton(), keyin(), redrawtextfield(), next_taborder();
static int flashns(), pulldown_deploy(), pulldown_hide(), flashpulldown(), redrawpulldown();
static int addpick(), buttonset(), addbuf(), delbuf(), setptsize(), measuretext(), savezone();


/* ===================================== */
/* GETGUI_INIT - Initialize.             */
/* ===================================== */
getgui_init( winname, rendermode )
char *winname;	   /* window name */
int rendermode;  /* 0 = autocrop / auto center on screen    1 = raw mode */
{
char buf[512];
autocrop = 0;

PLG_set_early_defaults();		/* set the early PLG defaults */

PLG_setoutlabel( winname );		/* set window name */
strcpy( PLG.progname, winname );	/* set program name (error messages, etc.) */

/* PLGX_flag( "initial_map", 0 ); */

if( rendermode == 0 ) {
	autocrop = 1;
	GL_make_unique_string( buf, 0 );		/* build a unique name for tmp file */
	sprintf( tmpfilename, "/tmp/getgui%s", buf );	
	PLG_setdumpfile( tmpfilename, "w" );		/* this will be the name for the pcode drawdump file */
	PLG_init( 'n' ); 				/* initialize PLG package (n=null device) */
	}
else PLG_init( 'x' );

setptsize( 10 );			
PLG.errfp = stderr;
PLG_backcolor( bgcolor );
PLG_color( "black" );

transx = transy = 0.0;
npick = 0;
return( 0 );
}


/* ================================ */
/* GETGUI_GO -  run it.             */
/* ================================ */

getgui_go( starttextfield, stopobj, stopobj2, stopobj3 )
int starttextfield;  /* if a textfield is to be active initially, this is it's objhandle (0 = no active textfield) */
int stopobj, stopobj2, stopobj3;       /* object ids that, when user picks/fills, causes this routine to return */
{
int i, e, ipick;
char objtype;
double x, y;
int firstingroup;

if( autocrop && PLG.dev == 'n' ) {
	render();  
	PLG_savewin();  /* save graphic so we can redraw on expose events (etc.) */
	PLG_flush();
	}

curtextfield = starttextfield;
if( curtextfield ) redrawtextfield( curtextfield, 1 );  /* start textfield as active.. */ 

pulldown_deployed = 0;


/* the main event loop */
while( 1 ) {
	getgui_getkey( &x, &y, &e ); 
	if( e == E_MESSAGE ) return( 0 );

	ipick = whichpick( x, y, &objtype );
	/* fprintf( stderr, "\n[pick %d %c]", ipick, objtype );  */

	if( e >= 1000 && ipick > 0 && objtype == 'T' ) {
		if( curtextfield && curtextfield != ipick ) redrawtextfield( curtextfield, 0 ); /* show current one as inactive.. */
		curtextfield = ipick;  /* textfield picked by mouse */
		}

	else if( e > 1000 && ( ipick == 0 || objtype != 'T' )) {  /* mouse click outside of current textfield */
		if( curtextfield ) redrawtextfield( curtextfield, 0 ); /* show current textfield as inactive.. */
		curtextfield = 0; 
		}

	else if( curtextfield && ( e == TAB || e == ENTER )) {	/* TAB or ENTER pressed while in textfield */
		int oldctf;
		redrawtextfield( curtextfield, 0 ); /* show current textfield as inactive.. */
		oldctf = curtextfield;
		curtextfield = next_taborder( curtextfield ); /* find next textfield in "tab order" (returns 0 if none) */
		if( curtextfield ) redrawtextfield( curtextfield, 1 ); /* show next textfield as active.. */
		if( oldctf == stopobj || oldctf == stopobj2 || oldctf == stopobj3 ) return( 0 );  /* we have use oldctf not curtextfield*/
		continue;
		}

	if( pulldown_deployed && ipick != pulldown_deployed && e >= 1000 ) {
		pulldown_hide( pulldown_deployed );
		pulldown_deployed = 0; 
		}

	if( curtextfield ) keyin( curtextfield, e, x, y );
	else if( objtype == BUTTON && ipick && e > 1000 ) flashbutton( ipick, 0 );
	else if( objtype == NAMESELECT && ipick && e > 1000 ) flashns( ipick, 0 );
	else if( objtype == PULLDOWN && ipick && e > 1000 ) { pulldown_deploy( ipick ); pulldown_deployed = ipick; }
	else if( objtype == PULLDOWN_CHOICE && ipick && e > 1000 ) flashpulldown( ipick );

	if( e > 1000 && objtype == BUTTON || objtype == PULLDOWN_CHOICE || objtype == NAMESELECT ) {
		/* find first in group.. */
		firstingroup = pick[ ipick-1 ].iprop2 +1;
		if( firstingroup == stopobj || firstingroup == stopobj2 || firstingroup == stopobj3 ) return( 0 );
		}
	}
return( 0 );
}

/* ===================================== */
/* RENDER - get the bounding box of the widgets, and render visibly with suitable      */
/* translation, so window is "cropped".  */
/* ===================================== */
static int
render()
{
double x, y;
int e;
char dev;
double xlo, ylo, xhi, yhi;
char buf[512];
FILE *fp;
int len, ix;
char op[20];

PLG_closedumpfile();	/* terminate writing to the dump file */


PLG_getbb( &xlo, &ylo, &xhi, &yhi ); /* get the current bounding box.. */

/* if fixed window size specified, compute window margins to produce fixed size */
if( winfixw > 0.0 && winfixh > 0.0 ) {
	winxmarg = (winfixw - (xhi-xlo))/2.0;
	/* if( winxmarg < 0.0 ) winxmarg = 0.0; */
	winymarg = (winfixh - (yhi-ylo))/2.0;
	/* if( winymarg < 0.0 ) winymarg = 0.0; */
	}

else if( xhi - xlo < 2.0 || yhi - ylo < 0.8 ) { winxmarg = winymarg = 0.5; } /* use a bigger margin for small content */

/* compute translation vector so that xlo,ylo is at 0.2,0.2 (using a winmarg of 0.2) */
transx = (xlo * -1.0) + winxmarg;
transy = (ylo * -1.0) + winymarg;


PLG_setsize( (xhi-xlo)+(winxmarg*2.0), (yhi-ylo)+(winymarg*2.0), win_x, win_y );  /* set size of window based on bounding box.. */
PLG_resetbb();  			/* re-initialize the bounding box (tho probably not really needed) */
PLG_init( 'x' ); 			/* start up x11.. */

/* render from the dump file, translating to 0,0 */
fp = fopen( tmpfilename, "r" );
if( fp == NULL ) { Eerr( 523, "Cannot open dump file", tmpfilename ); exit(1); }
PLG_backcolor( bgcolor );
PLG_clr(); /* clear the window to current background color */
while( fgets( buf, 512, fp ) != NULL ) {
	len = strlen( buf );
	buf[ len-1 ] = '\0';
	ix = 0;
	strcpy( op, GL_getok( buf, &ix ));
	x = atof( GL_getok( buf, &ix ));
	y = atof( GL_getok( buf, &ix ));
	if( buf[ix] == ' ' ) ix++;
	if( GL_member( op[0], "LMP" )) { 
		x += transx;
		y += transy;
		}
	PLG_pcode( op[0], x, y, &buf[ix] );
	}


fclose( fp );
unlink( tmpfilename );

return( 0 );
}

/* ===================================== */
/* SETPTSIZE - set the X11 point size and set our internal var charw to the      */
/* exact x11 char width.  Needed because some draw ops aren't done to 'x' but   */
/* to 'n' (null device).                  */
/* ===================================== */
static int 
setptsize( pt )
int pt;
{
double dpt;

if( pt < 5 ) pt = 5;
if( pt > 20 ) pt = 20;
/* PLG_textsize( pt ); */

/* do it this way to avoid wierd behavior */
dpt = pt;
PLG_pcode( 'I', dpt, 0.0, "" );
PLG.curtextsize = pt;
PLG.curtextheight = (dpt+2.0)/72.0;

charw = xcharwidths[ pt ];
PLG.curtextwidth = xcharwidths[ pt ];

return( 0 );
}


/* ===================================== */
/* WHICHPICK - look up an x, y location in the pick list.. provide info about */
/* what was clicked on.  Returns 0 if no object found, or the list location (first=1). */
/* ===================================== */
static int
whichpick( x, y, objtype )
double x, y;
char *objtype;
{
int i, firstingroup;

x -= transx; y -= transy;
for( i = npick-1; i >= 0; i-- ) {
	if( x >= pick[i].x1 && x <= pick[i].x2 &&
	    y >= pick[i].y1 && y <= pick[i].y2 ) {
		if( pick[i].objtype == PULLDOWN_DEPLOYED || pick[i].objtype == MISC ) continue; /* never need these */
		if( pulldown_deployed && pick[i].objtype == BUTTON ) continue; /* buttons may be rendered after pulldown but "underneath" */
		if( pick[i].objtype == PULLDOWN_CHOICE ) {
			/* check to see if active.. */
			firstingroup = pick[i].iprop2;
			if( pick[firstingroup].iprop == INACTIVE ) continue;
			}
		*objtype = pick[i].objtype;
		return( i+1 );
		}
	}
*objtype = '-';
return( 0 );
}


/* ===================================== */
/* GETGUI_LATE_REFRESH - triggered by a window expose.. redraw all transient content.*/
/* ===================================== */
getgui_late_refresh()
{
int i;
for( i = 0; i < npick; i++ ) {
	if( pick[i].objtype == BUTTON && pick[i].iprop ) {
		pick[i].iprop = 0; /* turn off, so it can be turned on again.. */
		flashbutton( i+1, 0 );
		}
	if( pick[i].objtype == NAMESELECT && pick[i].iprop ) {
		pick[i].iprop = 0; /* turn off, so it can be turned on again.. */
		flashns( i+1, 0 );
		}
	else if( pick[i].objtype == TEXTFIELD ) redrawtextfield( i+1, pick[i].iprop4 );
	else if( pick[i].objtype == PULLDOWN ) {
		if( pick[i].iprop == ACTIVE ) pick[i].iprop = INACTIVE; /* active pulldown becomes inactive */
		redrawpulldown( i+1 );
		}
	}
return( 0 );
}


/* ===================================== */
/* GETGUI_GETVAL - given a sethandle, copy the object's value into val. */
/* ===================================== */
getgui_getval( sethandle, val, method )
int sethandle;
char *val;
int method;
{
int i, j, nbuttons, lastbutton, first, iselect;
char tok[256];

strcpy( val, "" );
if( sethandle < 1 || sethandle > npick ) { fprintf( stderr, badlistlocmsg, "getval" ); return( 1 ); }
i = sethandle-1;

if( pick[i].objtype == TEXTFIELD ) strcpy( val, pick[i].objval );
else if( pick[i].objtype == BUTTON ) {
	nbuttons = pick[i].iprop3;
	lastbutton = i+nbuttons;
	for( ; i < lastbutton; i++ ) {
		if( pick[i].iprop ) {
			if( val[0] != '\0' ) strcat( val, listsep );
			if( method == FIRST_TOKEN ) { sscanf( pick[i].objval, "%s", tok ); strcat( val, tok ); }
			else if( method == INDEX_NUM ) { sprintf( tok, "%d",  (i-sethandle)+2 ); strcat( val, tok ); }
			else strcat( val, pick[i].objval ); 
			}
		}
	}
else if( pick[i].objtype == NAMESELECT ) {
	nbuttons = pick[i].iprop3;
	lastbutton = i+nbuttons;
	for( ; i < lastbutton; i++ ) {
		if( pick[i].iprop ) {
			if( val[0] != '\0' ) strcat( val, listsep );
			if( method == FIRST_TOKEN ) { sscanf( pick[i].objval, "%s", tok ); strcat( val, tok ); }
			else if( method == INDEX_NUM ) { sprintf( tok, "%d", (i-sethandle)+2 ); strcat( val, tok ); }
			else strcat( val, pick[i].objval ); 
			}
		}
	}

else if( pick[i].objtype == PULLDOWN ) {
	if( method == FIRST_TOKEN ) sscanf( pick[i].objval, "%s", val );
	else if( method == INDEX_NUM ) { 
		for( j = i+2; j < i+pick[i].iprop3+2; j++ ) {
			if( strcmp( pick[i].objval, pick[j].objval )==0 ) { sprintf( val, "%d", j-(i+1) ); break; }
			}
		}
	else strcpy( val, pick[i].objval );
	}
	
else { fprintf( stderr, "getval: this type (%d) doesn't have a usable value\n", pick[i].objtype ); return( 1 ); }

return( 0 );
}


/* ===================================== */
/* BUTTONS - Initial rendition of a button set with optional multiple tiers. */
/* Returns a sethandle (1+) or 0 on error */
/* ===================================== */

getgui_buttons( x, y, align, choicelist, ptsize, colorlist, wraplen, choosemode )
double x, y;		/* location (left edge, vertical center of top tier) */
char *align;		/* center=buttons centered on x,y  left=left btns edge at x,y  right=right btns edge at x,y */
char *choicelist;	/* bar-delimited, eg. "Continue|OK" */
int ptsize;		/* point size for button label (also affects overall button size) */
char *colorlist;	/* bar-delmited, eg. "red|blue|rgb(0.4,0.5,0.7)"   If not enough members to cover all buttons, 
				last member is repeated.  Thus, one color may be given for all buttons.
				Same colorlist is used for each button tier. */
int wraplen;		/* if specified > 0, buttons will wrap to next row when char width passes this */
char *choosemode;	/* "single" (single choice) or "multiple" (multiple choice) */
{
int buflen, ix;
int sethandle;
int i, bcount;
double twidth;
char buf[512];
double bmarg;
int zoneid;

sethandle = npick+1;

if( strcmp( colorlist, "multi" )==0 ) colorlist = bcolorlist; /* a canned color list */

PLG_init_bb2();   /* initialize an object bounding box in case app wants to get zone of entire box later */

if( relalign == 2 || relalign == 3 ) x = x_override; 
if( relalign == 1 || relalign == 3  ) y = y_override; 
relalign = 0;

/* count # buttons */
for( i = 0, bcount = 1; choicelist[i] != '\0'; i++ ) { if( choicelist[i] == '|' ) bcount++; }
pick[ npick ].iprop3 = bcount; /* remember # of buttons in first button pick list entry.. */
pick[ npick ].iprop4 = choosemode[0];

bmarg = ( ptsize * 0.01 ) - 0.04;

if( wraplen > 0 && strlen( choicelist ) > wraplen ) {
	/* generate multiple tiers of buttons */
	strcpy( buf, "" ); buflen = 0;
	ix = 0;
	while( 1 ) {
		GL_getchunk( &buf[ buflen ], choicelist, &ix, "|" );

		if( buf[buflen] == '\0' && buflen > 0 ) { /* display remaining buttons */
			buf[ buflen-1 ] = '\0'; /* remove last | */
			buttonset( x, y, align, buf, ptsize, colorlist, sethandle-1 );
			break;
			}
		buflen += strlen( &buf[buflen] );
		buf[ buflen++ ] = '|'; buf[ buflen ] = '\0';
		if( buflen > wraplen ) {  /* wrap encountered.. display a tier of buttons.. */
			buf[ buflen-1 ] = '\0'; /* remove last | */
			buttonset( x, y, align, buf, ptsize, colorlist, sethandle-1 );	
			buflen = 0;
			y -= PLG.curtextheight + bmarg + bmarg + (bmarg*bbevelsize*2.0);
			}
		}
	}
else buttonset( x, y, align, choicelist, ptsize, colorlist, sethandle-1 );

zoneid = savezone(); /* add a pick list entry for entire */

return( sethandle );
}

/* ----------------------------------------------------------------- */
/* BUTTONSET - handle everything about the button set except multiple tiers, controlled above */

static int 
buttonset( x, y, align, choicelist, ptsize, colorlist, firstingroup )
double x, y;					
char *align, *choicelist; 			/* all args - same as for getgui_buttons() above */	
int ptsize;
char *colorlist;
int firstingroup;
{
int i, j;
int bcount;
double twidth;
char label[80];
char color[40], colorhint[40];
double bheight, bwidth, x0, bmarg;
int ix1, ix2;

if( ptsize == 0 ) ptsize = btextsize;
setptsize( ptsize );
bmarg = ( ptsize * 0.01 ) - 0.04;


/* count # of buttons in this tier.. */
for( i = 0, bcount = 1; choicelist[i] != '\0'; i++ ) { if( choicelist[i] == '|' ) bcount++; }
if( bcount < 1 ) return( 0 );

bheight = PLG.curtextheight + bmarg + bmarg;

/* compute width of button set */
twidth = (double) (strlen( choicelist )) * charw;
twidth += ((double)bcount * (bmarg*2)) + ((double)(bcount-1) * bsep );

/* determine left side of button set based on alignment and twidth.. */
if( align[0] == 'r' ) x0 = x - twidth;
else if( align[0] == 'c' ) x0 = x - (twidth/2.0);
else x0 = x;

strcpy( colorhint, BUTTON_COLOR );

/* parse choicelist, colorlist, and render buttons.. */
ix1 = ix2 = 0;
x = x0; y -= ((PLG.curtextheight/2.0) + bmarg );
while( 1 ) {
	GL_getchunk( label, choicelist, &ix1, "|" );
	if( label[0] == '\0' ) break;
	GL_getchunk( color, colorlist, &ix2, "|" );
	if( color[0] == '\0' ) strcpy( color, colorhint );
	else strcpy( colorhint, color );
	
	/* x, y are now lower left corner of button */

	bwidth = (strlen( label ) * charw) + bmarg + bmarg;
	PLG_cblock( x, y, x+bwidth, y+bheight, color, (bdressup)?0:1 );
	if( bdressup == 1 ) PLG_cblockdress( x, y, x+bwidth, y+bheight, bmarg*bbevelsize, "gray(0.7)", "gray(0.95)", 0.0, "" );
	else if( bdressup == 2 ) PLG_cblockdress( x, y, x+bwidth, y+bheight, 0.0, "", "", bmarg*bbevelsize, "gray(0.1)" );

	/* add the button to pick list */
	if( npick >= MAXPICK-1 ) { fprintf( stderr, fullobjmsg ); return( 0 ); }
	pick[ npick ].objtype = BUTTON; strcpy( pick[ npick ].objval, label );
	pick[ npick ].x1 = x; pick[ npick ].y1 = y; pick[ npick ].x2 = x+bwidth; pick[ npick ].y2 = y+bheight;
	strcpy( pick[ npick ].cprop, color );
	pick[ npick ].iprop = 0;
	pick[ npick ].iprop2 = firstingroup;
	pick[ npick ].iprop4 = ptsize;
	npick++;

	PLG_color( "black" );
	PLG_mov( x + (bwidth/2.0), y+bmarg+(PLG.curtextheight*0.15) );  /* 0.15 = move text up a little */
	PLG_dotext( label, 'C' );

	x += (bwidth + bsep);
	if( bdressup ) x += (bmarg*bbevelsize)*2.0;
	}
return( 0 );
}


/* ========================================  */
/* FLASHBUTTON - handle a click on a button. */
/* ========================================  */
static int
flashbutton( listloc, recurs )
int listloc;  /* location in pick list of item (first=1) */
int recurs;
{
int  j; 
double x1, y1, x2, y2;
char color[40];
int ipk, firstingroup, lastingroup, ptsize;
double bmarg;

if( listloc < 1 || listloc > npick ) { fprintf( stderr, badlistlocmsg, "flashbutton" ); return( 1 ); }
ipk = listloc-1;
if( pick[ipk].objtype != BUTTON ) { fprintf( stderr, "obj %d is not a button\n", listloc ); return( 1 ); }

x1 = pick[ipk].x1 + transx; y1 = pick[ipk].y1 + transy; x2 = pick[ipk].x2 + transx; y2 = pick[ipk].y2 + transy;

ptsize = pick[ipk].iprop4;
setptsize( ptsize );
bmarg = ( ptsize * 0.01 ) - 0.04;

if( pick[ipk].iprop == 0 ) { strcpy( color, bflashcolor ); pick[ipk].iprop = 1; } /* select */
else { strcpy( color, pick[ipk].cprop ); pick[ipk].iprop = 0; } /* unselect */
PLG_cblock( x1, y1, x2, y2, color, (bdressup)?0:1 );
PLG_mov( x1 + (( x2 - x1 )/2.0), y1+bmarg+(PLG.curtextheight*0.15) ); /* 0.15 = move text up a little */
PLG_color( "black" );
PLG_dotext( pick[ipk].objval, 'C' );

if( recurs ) return( 0 );

/* find first button in group.. */
firstingroup = pick[ipk].iprop2;
if( pick[ firstingroup ].iprop4 == 'm' ) return( 0 ); /* multiple choice, don't do radio action.. */

/* see if any other name selected, and un-select it.. */
lastingroup = firstingroup + pick[ firstingroup ].iprop3;
for( j = firstingroup; j < lastingroup; j++ ) {
	if( j == ipk ) continue;
	if( pick[j].iprop ) {
		flashbutton( j+1, 1 );  /* recursive call */
		break;
		}
	}

PLG_flush();
usleep( FLASH_uS );

return( 0 );
}



/* ===================================== */
/* TEXTFIELD - initial rendition of a text field. */
/* Returns a sethandle (1+) or 0 on error */
/* ===================================== */

getgui_textfield( x, y, align, length, maxlen, initval, initcursor, modestr )
double x, y;	/* location (left edge, vertical middle of field) */
char *align;	/* left, center, right */
int length;	/* displayable length (# chars)
int maxlen;	/* maximum allowable length (#chars) */
char *initval;	/* text to initially appear in the field */
int initcursor;	/* where cursor should be placed initially: use 0 for leftmost; use eg. 99 to place at end of initval */
char *modestr;	/* "P" for password mode */
{
int i;
double x0;
double twidth, theight;


setptsize( ttextsize );
PLG_color( "black" );

if( relalign == 2 || relalign == 3 ) x = x_override;
if( relalign == 1 || relalign == 3 ) y = y_override; 
relalign = 0;

theight = (PLG.curtextheight*0.7) + tmarg + tmarg;

/* compute width of textfield */
twidth = (double) length * charw;
twidth += (tmarg * 2); 

/* determine left side of button set based on alignment and twidth.. */
if( align[0] == 'r' ) x0 = x - twidth;
else if( align[0] == 'c' ) x0 = x - (twidth/2.0);
else x0 = x;

x = x0; y -= ((PLG.curtextheight/2.0) + tmarg );

if( GL_member( 'P', modestr ) && ( maxlen == 0 || maxlen > length )) maxlen = length; /* password fields shouldn't scroll */


/* add the textfield to pick list */
if( npick >= MAXPICK-1 ) { fprintf( stderr, fullobjmsg ); return( 0 ); }
pick[ npick ].objtype = TEXTFIELD; strcpy( pick[ npick ].objval, initval );
pick[ npick ].x1 = x; pick[ npick ].y1 = y; pick[ npick ].x2 = x+twidth; pick[ npick ].y2 = y+theight;
if( initcursor > strlen( initval ) ) pick[ npick ].iprop = strlen( initval ); 
else pick[ npick ].iprop = initcursor;
pick[ npick ].iprop2 = length;
pick[ npick ].iprop3 = maxlen; pick[ npick ].iprop4 = 0;
strcpy( pick[ npick ].cprop, modestr );

npick++;

redrawtextfield( npick, 0 );   /* draw textfield contents */

return( npick );  
}


/* ====================================== */
/* KEY_IN - handle keystrokes while in a text field.  */
/* ====================================== */
static int
keyin( listloc, key, mousex, mousey )
int listloc;  /* location in pick list of item (first=1) */
int key;
double mousex, mousey;
{
int i, ccol, len;
double xleft;
int maxlen;

if( key == TAB || key == ENTER ) return( 0 );   /* don't accept Tab or Enter */
if( listloc < 1 || listloc > npick ) { fprintf( stderr, badlistlocmsg, "keyin" ); return( 1 ); }
if( pick[listloc-1].objtype != TEXTFIELD ) { fprintf( stderr, "obj %d is not a textfield\n", listloc ); return( 1 ); }

listloc--; /* adjust for off-by-one */

ccol = pick[ listloc ].iprop;
len = strlen( pick[ listloc ].objval );
xleft = pick[ listloc ].x1 + transx;
maxlen = pick[ listloc ].iprop3;
if( maxlen == 0 ) maxlen = MAXVALLEN;

if( key == LEFT && ccol > 0 ) ccol--;
else if( key == RIGHT && ccol < len ) ccol++;
else if( key == HOME || key == CTRL_A ) ccol = 0;
else if( key == END || key == CTRL_E ) ccol = len;
else if( key == E_MOUSE_LEFT ) { /* move cursor to desired location */
	if( mousex > xleft + ( len * charw)) ccol = len;
	else ccol = (int)((mousex - xleft) / charw);
	}
else if( key == CTRL_K ) pick[ listloc ].objval[ ccol ] = '\0';
else if( key == CTRL_U ) { pick[ listloc ].objval[0] = '\0'; ccol = 0; }
else if( key == CTRL_D || key == DEL || key == BACKSPACE ) {
	if( key == BACKSPACE || (key == DEL && DEL_EQUIV_BACKSPACE) ) { 
		if( ccol > 0 ) ccol--; 
		else if( ccol == 0 ) goto BOTTOM;
		}
	if( ccol < len ) delbuf( pick[ listloc ].objval, ccol, 1 );
	}
else if( key >= 32 && key <= 126 && ccol < (maxlen) ) addbuf( pick[ listloc ].objval, ccol++, key );  /* typed characters.. */

#ifdef NOTYET
else if( key == E_MOUSE_MIDDLE || key == E_MOUSE_RIGHT ) { /* increment/decrement */
	if( ey < y || ey > (ey + height) ) break;
	else if( ex < x || ex > x + width ) break;
	else if( ex > x + (strlen(str) * charw )) continue;
	j = (int)(((ex - x) / charw );
	if( isdigit( str[j] )) {
		if( key == E_MOUSE_MIDDLE ) str[j]--;
		else if( key == E_MOUSE_RIGHT ) str[j]++;
		if( str[j] > '9' ) str[j] = '0';
		if( str[j] < '0' ) str[j] = '9';
		}
	}
#endif


BOTTOM:
pick[ listloc ].iprop = ccol;

redrawtextfield( listloc+1, 1 ); /* +1 because of off-by-one */
return( 0 );
}

/* -------------------------------------------- */
/* DELBUF - Delete some text from the buffer.  txt = text buffer;
/* itxt = current position; dels = number of chars to delete */

static int delbuf( txt, itxt, dels )
char *txt;
int itxt;
int dels;
{
int i;
if( itxt < 0 || itxt > strlen( txt ) ) fprintf( stderr, "delbuf: itxt (%d) out of range\n", itxt );
for( i = itxt; i <= (int) (strlen( txt )) - (int) dels; i++ ) {
	txt[i] = txt[i+dels];
	}
}

/* -------------------------------------------- */
/* ADDBUF - Add some text to the buffer.  txt = text buffer;
/* itxt = current position; newchar = char to be added */

static int addbuf( txt, itxt, newchar )
char *txt;
int itxt;
char newchar;
{
int i;
if( itxt < 0 || itxt > strlen( txt ) ) fprintf( stderr, "addbuf: itxt (%d) out of range\n", itxt );
for( i = strlen( txt ); i >= itxt ; i-- ) {
	txt[i + 1] = txt[i];
	}
txt[itxt] = newchar;
}

/* ===================================== */
/* REDRAWTEXTFIELD - redraw current contents of a textarea (text & cursor) */
/* ===================================== */
static int
redrawtextfield( listloc, active )
int listloc;
int active;
{
int j, len, frameadj, ccol, ipk;
double x, y, x2, y2;
char *str;
char tmpchar;
int pwmode;
char starstr[ MAXVALLEN+1 ];

if( listloc < 1 || listloc > npick ) { fprintf( stderr, badlistlocmsg, "redrawtextfield" ); return( 1 ); }
if( pick[listloc-1].objtype != TEXTFIELD ) { fprintf( stderr, "obj %d is not a textfield\n", listloc ); return( 1 ); }

ipk = listloc-1;
x = pick[ ipk ].x1 + transx;
y = pick[ ipk ].y1 + transy;
x2 = pick[ ipk ].x2 + transx;
y2 = pick[ ipk ].y2 + transy;

ccol = pick[ ipk ].iprop;
len = pick[ ipk ].iprop2;
pick[ ipk ].iprop4 = active;

if( GL_member( 'P', pick[ ipk ].cprop )) pwmode = 1;
else pwmode = 0;

str = pick[ ipk ].objval;

if( ccol > len ) frameadj = ccol - len;
else frameadj = 0;


PLG_cblock( x, y, x2, y2, tcolor, 1 );

/* outline */
PLG_color( "gray(0.8)" );
PLG_linetype( 0, 1.0, 1.0 );
PLG_mov( x, y ); PLG_lin( x2, y ); PLG_lin( x2, y2 ); 
PLG_color( "black" );
PLG_lin( x, y2 ); PLG_lin( x, y ); 
PLG_linetype( 0, 0.5, 1.0 );

setptsize( ttextsize );
if( pick[ipk].objval[0] != '\0' ) {
	if( pwmode ) { 
		for( j = 0; str[j] != '\0'; j++ ) starstr[j] = '*';
		starstr[j] = '\0';
		str = starstr;
		}
	PLG_mov( x+tmarg, y+tmarg );
	tmpchar = str[ frameadj + len ];
	str[ frameadj + len ] = '\0'; /* truncate to displayable length */
	PLG_dotext( &str[ frameadj ], 'L' );
	str[ frameadj + len ] = tmpchar; /* restore */
	}

/* position and draw the cursor.. */
if( active ) {
	x += tmarg + (charw * (ccol-frameadj) );
	y += 0.05;
	cursorx = x; cursory = y; /* remember location so we can blink it later.. */
	PLG_mov( x-0.04, y ); PLG_lin( x+0.04, y ); PLG_mov( x, y );
	y += PLG.curtextheight * 1.1;
	cursortopy = y;
	PLG_lin( x, y ); PLG_mov( x-0.04, y ); PLG_lin( x+0.04, y );
	}

return( 0 );
}

/* ==================================== */
/* NEXT_TABORDER - return handle for next textfield in list.  Return 0 if */
/* only one textfield, or if at end.	*/
/* ==================================== */
static int
next_taborder( oldtf )
int oldtf; /* listloc of current text field (first=1) */
{
int i;

if( oldtf < 1 || oldtf > npick ) { fprintf( stderr, badlistlocmsg, "next_taborder" ); return( 1 ); }
if( pick[oldtf-1].objtype != TEXTFIELD ) { fprintf( stderr, "obj %d is not a textfield\n", oldtf ); return( 1 ); }

/* note that oldtf is (1+).. but start with it since we don't want to examine current textfield's entry */
for( i = oldtf; i < npick; i++ ) {
	if( pick[i].objtype == 'T' ) return( i+1 );
	}
return( 0 );
}


/* ===================================== */
/* NAMESELECT - initial rendition of a  nameselect box.  This code also does  */
/* double duty for pulldowns. (pulldown=1) */
/* ===================================== */
getgui_nameselect( x, y, align, valign, inbuf, delimmode, initval, choosemode, minwidth, pickrep )
double x, y;	/* upper left corner */
char *align;	/* how to align box against the given 'x' coordinate, either "left", "center", or "right" */
char *valign;   /* how to align box against the given 'y' coordinate, either "top", "center", or "bottom" */
char *inbuf;		/* !command or text rows w/ newlines */
char *delimmode;	/* either "word", "chunk", or "line".  "chunk" is like word but requires > 1 space char. */
char *initval;		/* initially selected value */
char *choosemode;	/* "single" (single choice), "multiple" (multiple choice), or "pulldown" (special pulldown mode) */
int minwidth;		/* if > 0, select box will be at least minwidth characters wide */
int pickrep;		/* see #defines */
{
int i, maxline, nlines, len, sethandle, ix;
double nwidth, nheight, x0;
FILE *pfp;
double fldleft, fldrt;
int infld, fldstart;
char buf[512], buf2[512];
int nflds, firstpick;
int linemode, addit;
int pulldown;  /* if 1, add entries to pick list but don't display them yet.. */
int iln;

sethandle = npick+1;

if( delimmode[0] == 'l' ) linemode = 1;
else linemode = 0;

if( choosemode[0] == 'p' ) pulldown = 1; 
else pulldown = 0;

if( !linemode && pickrep == FIRST_TOKEN ) pickrep = ENTIRE_TEXT;

/* set text size */
setptsize( ntextsize );

if( relalign == 2 || relalign == 3) x = x_override;
if( relalign == 1 || relalign == 3) y = y_override;
if( !pulldown ) relalign = 0;

if( !pulldown ) PLG_init_bb2();   /* initialize an object bounding box in case app wants to get zone of entire box later */

/* get the dimensions of the text.. */
maxline = 0; nlines = 0; 
if( inbuf[0] == '!' ) pfp = popen( &inbuf[1], "r" );
else ix = 0;
while( 1 ) {
	if( inbuf[0] == '!' ) { if( fgets( buf2, 511, pfp ) == NULL ) break; }
	else { GL_getchunk( buf2, inbuf, &ix, "\n" ); if( buf2[0] == '\0' ) break; }
	GL_expand_tabs( buf, buf2 );
	len = strlen( buf );
	if( len > maxline ) maxline = len;
	nlines++;
	}
if( inbuf[0] == '!' ) pclose( pfp );

/* sanity bounds */
if( nlines > MAXSELECTROWS ) nlines = MAXSELECTROWS;
if( maxline > MAXSELECTROWLEN ) maxline = MAXSELECTROWLEN;
if( nlines < 1 ) fprintf( stderr, "nameselect: warning, no content to display\n" );
if( minwidth > 0 && maxline < minwidth ) maxline = minwidth;
nwidth = (maxline * charw) + nmarg + nmarg;
nheight = (nlines * PLG.curtextheight) + nmarg + nmarg;

/* determine location of left side, taking into account align.. */
if( align[0] == 'r' ) x0 = x - nwidth;
else if( align[0] == 'c' ) x0 = x - (nwidth/2.0);
else x0 = x;

/* determine location of top, taking into account valign.. */
if( valign[0] == 'b' ) y += nheight;
else if( valign[0] == 'c' ) y += (nheight/2.0);


if( pulldown ) firstpick = npick-1;
else 	{
	firstpick = npick;
	PLG_cblock( x0, y-nheight, x0+nwidth, y, ncolor, 0 );/* render backing box */
	PLG_color( "black" );
	PLG_linetype( 0, 1.0, 1.0 );
	PLG_mov( x0, y-nheight ); PLG_lin( x0, y ); PLG_lin( x0+nwidth, y );
	PLG_linetype( 0, 0.5, 1.0 );
	}

if( pulldown ) {
	/* save deployed zone dimensions.. */
	pick[ npick ].objtype = PULLDOWN_DEPLOYED;
        pick[ npick ].x1 = x0; pick[ npick ].y1 = y-nheight;
        pick[ npick ].x2 = x0+nwidth; pick[ npick ].y2 = y;
        pick[ npick ].iprop2 = npick-1;
        npick++;
	/* also update bounding box.. since backing box wasn't drawn.. */
	PLG_bb( x0, y-nheight );
	PLG_bb( x0+nwidth, y );
	}


nflds = 0;
x0 += nmarg;
y -= PLG.curtextheight;

if( inbuf[0] == '!' ) pfp = popen( &inbuf[1], "r" );
else ix = 0;

/* render text */
iln = 0;
while( 1 ) {
	if( inbuf[0] == '!' ) { 
		if( fgets( buf2, 511, pfp ) == NULL ) break; 
		}
	else 	{ 
		GL_getchunk( buf2, inbuf, &ix, "\n" ); 
		if( buf2[0] == '\0' ) break; 
		strcat( buf2, "\n" ); /* put the newline back on.. needed below */
		}
	GL_expand_tabs( buf, buf2 );

	buf[ MAXSELECTROWLEN ] = '\0'; 
	if( iln >= MAXSELECTROWS ) break;

	if( !pulldown ) { 
		PLG_mov( x0, y );
		PLG_dotext( buf, 'L' );
		} 
		
	/* parse whitespace-delimited fields (or entire line).. */
	infld = 0;
	for( i = 0, addit = 0; buf[i] != '\0'; i++ ) {

		if( linemode && i == 0 ) { fldleft = x0; fldstart = i; }
		else if( linemode && buf[i] == '\n' ) { fldrt = x0 + nwidth - (nmarg*1.8); addit = 1; }
		else if( !linemode && !infld && !isspace( buf[i] )) 
			{ fldleft = x0 + ( i * charw); fldstart = i; infld = 1; } /* begin word or chunk */
		else if( delimmode[0] == 'w' && infld && ( isspace( buf[i] ) || buf[i] == '\0' )) 
			{ fldrt = x0 + ( i * charw); infld = 0; addit = 1; } /* end word */
		else if( delimmode[0] == 'c' && infld && isspace( buf[i] ) && isspace( buf[i+1] ))
			{ fldrt = x0 + ( i * charw); infld = 0; addit = 1; } /* end chunk */
		else if( delimmode[0] == 'c' && infld && buf[i] == '\0' || buf[i+1] == '\0' )  
			{ fldrt = x0 + ( i * charw); infld = 0; addit = 1; } /* end chunk - special case for end of line */

		if( addit ) {
			/* add entry to pick list.. */
			if( npick >= MAXPICK-20 ) { fprintf( stderr, fullobjmsg ); return( 0 ); } 
			/* note, MAXPICK-20 since this is most likely place of overflow occurance, ie. big ls(1) result */
			/* .. allows a few add'l objects */
			if( pulldown ) pick[ npick ].objtype = PULLDOWN_CHOICE;
			else pick[ npick ].objtype = NAMESELECT; 
			buf[i] = '\0';
			if( buf[i-1] == ' ' ) buf[i-1] = '\0'; /* handle trailing space artifact of convertnl() */
			strcpy( pick[ npick ].objval, &buf[fldstart] );
			pick[ npick ].x1 = fldleft; pick[ npick ].y1 = y-(PLG.curtextheight*0.1);  /* note0001 */
			pick[ npick ].x2 = fldrt; pick[ npick ].y2 = y+(PLG.curtextheight*0.68);
			pick[ npick ].iprop2 = firstpick;
			pick[ npick ].iprop = 0;
			npick++;
			if( initval[0] != '\0' && (
			  ( pickrep == ENTIRE_TEXT && strcmp( pick[ npick-1 ].objval, initval )==0 ) || 
			  ( pickrep == FIRST_TOKEN && strncmp( initval, buf, strlen(initval) )==0 ) ||
			  ( pickrep == INDEX_NUM && atoi( initval ) == (npick-firstpick) ) 
			  ) ) {
				if( !pulldown ) flashns( npick, 0 ); /* display it as selected */
				else pick[ npick-1 ].iprop = 1;
				}
			nflds++;
			addit = 0;
			}
		}
	y -= PLG.curtextheight;
	}
if( inbuf[0] == '!' ) pclose( pfp );

/* remember how many cells in this nameselect.. */
pick[ firstpick ].iprop3 = nflds;
if( pulldown ) pick[ firstpick ].iprop4 = 's'; /* pulldowns always single choice */
else pick[ firstpick ].iprop4 = choosemode[0];

savezone(); /* add a pick list entry for entire */

return( sethandle );
}

/* ========================================  */
/* FLASHNS - handle a click on a nameselect box. */
/* ========================================  */

flashns( listloc, recurs )
int listloc;  /* location in pick list of item (first=1) */
int recurs;
{
int j, firstingroup, lastingroup, ipk;
double x1, y1, x2, y2;
char color[40];
double selmarg;

if( listloc < 1 || listloc > npick ) { fprintf( stderr, badlistlocmsg, "flashns" ); return( 1 ); }
if( pick[listloc-1].objtype != NAMESELECT && 
  pick[ listloc-1].objtype != PULLDOWN_CHOICE ) { fprintf( stderr, "obj %d is not a nameselect\n", listloc ); return( 1 ); }
ipk = listloc-1;

x1 = pick[ipk].x1 + transx; y1 = pick[ipk].y1 + transy; x2 = pick[ipk].x2 + transx; y2 = pick[ipk].y2 + transy;

setptsize( ntextsize );
selmarg = charw * 0.2;

if( pick[ipk].iprop == 0 ) { strcpy( color, nflashcolor ); pick[ipk].iprop = 1; } /* select */
else { strcpy( color, ncolor ); pick[ipk].iprop = 0; } /* unselect */
PLG_cblock( x1-selmarg, y1-selmarg, x2+selmarg, y2+selmarg, color, 0 );

if( pick[ipk].iprop == 1 ) PLG_color( "white" );
else PLG_color( "black" );
PLG_mov( x1, y1+(PLG.curtextheight*0.1) );   /* adjustment amount (0.1) must match note0001 above */
PLG_dotext( pick[ipk].objval, 'L' );
PLG_color( "black" );

if( recurs ) return( 0 );

/* now see if any other name selected, and un-select it.. */
firstingroup = pick[ipk].iprop2;
if( pick[ firstingroup ].iprop4 == 'm' ) return( 0 ); /* multi-choice, don't do radio action */

lastingroup = pick[ firstingroup ].iprop3;
if( pick[ firstingroup ].objtype == PULLDOWN ) firstingroup += 2; /* skip 1st 2 pulldown entries.. */
lastingroup += firstingroup;

for( j = firstingroup; j < lastingroup; j++ ) {
	if( j == ipk ) continue;
	if( pick[j].iprop ) {
		flashns( j+1, 1 );  /* recursive call */
		break;
		}
	}

return( 0 );
}


/* =============================================== */
/* PULLDOWN - initial rendition of a pulldown menu bar */
/* Returns the handle for the newly created object. */
/* =============================================== */

getgui_pulldown( x, y, align, valign, inbuf, initval, pickrep )
double x, y;
char *align;
char *valign; /* vertical alignment for the deployed pulldown zone, relative to the pulldown bar */
char *inbuf;
char *initval;
int pickrep; /* see #defines */
{
int i, j, sethandle, ipk;
double x1, y1, x2, y2, depy;

sethandle = npick+1;
if( npick >= MAXPICK-1 ) { fprintf( stderr, fullobjmsg ); return( 0 ); }

npick++;  /* advance over first pulldown entry.. we'll come back and fill it in later.. */

/* don't do relalign here.. X will be done by nameselect.. Y done afterwards (below) */

/* set up the (not yet visible) pulldown choices now, to get max width.. */
/* this will add a pick entry for the entire deployed zone, as well as each choice. */
if( valign[0] == 't' ) depy = y-PLG.curtextheight;
else depy = y;
getgui_nameselect( x, depy, align, valign, inbuf, "line", initval, "pulldown", 0, pickrep );

setptsize( ptextsize );

if( relalign == 1 || relalign == 3) y = y_override;
relalign = 0;
y -= ((PLG.curtextheight*0.5) + nmarg ); 

/* now go back and save the properties for the closed pulldown bar... */
ipk = sethandle - 1;
pick[ ipk ].objtype = PULLDOWN;
strcpy( pick[ ipk ].objval, initval );
/* x locations will be same as deployed zone.. copy from there.. */
x1 = pick[ ipk+1 ].x1; y1 = y-nmarg; 
x2 = pick[ ipk+1 ].x2 + 0.2; y2 = y+(PLG.curtextheight*0.7)+nmarg;   /* x2 includes extra room for the little triangle.. */
pick[ ipk ].iprop = INACTIVE; 
/* note: iprop2 has been set to choice mode */
/* note: iprop3 has already been set to number in group */
pick[ ipk ].x1 = x1; pick[ ipk ].y1 = y1; pick[ ipk ].x2 = x2; pick[ ipk ].y2 = y2;

redrawpulldown( sethandle );

return( sethandle );
}

/* =========================== */
/* PULLDOWN_DEPLOY - Deploy the given pulldown.  */
/* =========================== */
static int
pulldown_deploy( listloc )
int listloc; /* a PULLDOWN picklist entry (first=1) */
{
int ipk, j;
int firstingroup, lastingroup;

if( listloc < 1 || listloc > npick ) { fprintf( stderr, badlistlocmsg, "pdd" ); return( 1 ); }
if( pick[listloc-1].objtype != PULLDOWN ) { fprintf( stderr, "obj %d is not a pulldown\n", listloc ); return( 1 ); }
ipk = listloc;

/* ipk indexes the PULLDOWN_DEPLOYED entry.. */
PLG_cblock( pick[ipk].x1+transx, pick[ipk].y1+transy, pick[ipk].x2+transx, pick[ipk].y2+transy, "white", 1 );
pick[ ipk-1 ].iprop = ACTIVE;
firstingroup = ipk+1;
lastingroup = firstingroup + pick[ipk-1].iprop3;
for( j = firstingroup; j < lastingroup; j++ ) {
	if( pick[ j ].iprop == 0 ) pick[ j ].iprop = 1; /* select so it will be shown as unselected.. */
	else pick[ j ].iprop = 0;

	flashns( j+1, 1 );
	}

return( 0 );
}

/* ======================= */
/* PULLDOWN_HIDE - Hide the given pulldown */
/* ======================= */
static int
pulldown_hide( listloc )
int listloc; /* a PULLDOWN picklist entry (first=1) */
{
if( listloc < 1 || listloc > npick ) { fprintf( stderr, badlistlocmsg, "pdh" ); return( 1 ); }
if( pick[listloc-1].objtype != PULLDOWN ) { fprintf( stderr, "obj %d is not a pulldown\n", listloc ); return( 1 ); }
PLG_restorewin();
getgui_late_refresh();
pick[ listloc-1 ].iprop = INACTIVE;
return( 0 );
}

/* =============================== */
/* FLASHPULLDOWN - Get a user choice on a pulldown */
/* =============================== */
static int
flashpulldown( listloc )
int listloc;  /* a PULLDOWN_CHOICE picklist entry (first=1) */
{
int firstingroup;
if( listloc < 1 || listloc > npick ) { fprintf( stderr, badlistlocmsg, "fpd" ); return( 1 ); }
if( pick[listloc-1].objtype != PULLDOWN_CHOICE ) { fprintf( stderr, "obj %d is not a pulldown choice\n", listloc ); return( 1 ); }
flashns( listloc, 0 );
firstingroup = pick[ listloc-1 ].iprop2;
strcpy( pick[ firstingroup ].objval, pick[ listloc-1 ].objval ); /* choice value becomes pulldown value */
PLG_flush();
usleep( FLASH_uS );
pulldown_hide( firstingroup+1 );
redrawpulldown( firstingroup+1 );
return( 0 );
}

/* ============================== */
/* REDRAWPULLDOWN - Redraw the pulldown bar, showing current value.   */
/* ============================== */
static int
redrawpulldown( listloc )
int listloc; /* a PULLDOWN picklist entry (first=1) */
{
int ipk;
double x1, y1, x2, y2;

if( listloc < 1 || listloc > npick ) { fprintf( stderr, badlistlocmsg, "rdpd" ); return( 1 ); }
if( pick[listloc-1].objtype != PULLDOWN ) { fprintf( stderr, "obj %d is not a pulldown\n", listloc ); return( 1 ); }
ipk = listloc-1;
x1 = pick[ ipk ].x1 + transx; y1 = pick[ ipk ].y1 + transy; x2 = pick[ ipk ].x2 + transx; y2 = pick[ ipk ].y2 + transy;

PLG_cblock( x1, y1, x2, y2, pcolor, 0 );

/* the little trangle area.. */
PLG_cblock( x2-0.22, y1+0.02, x2-0.02, y2-0.04, "gray(0.8)", 0 );
PLG_cblockdress( x2-0.22, y1+0.02, x2-0.02, y2-0.04, 0.02, "black", "gray(0.95)", 0.0, "" );
PLG_color( "black" );
PLG_mov( x2-0.18, y2-0.1 ); PLG_path( x2-0.08, y2-0.1 ); PLG_path( x2-0.13, y2-0.17 ); PLG_path( x2-0.18, y2-0.1 );
PLG_colorfill( "black" );

PLG_color( "black" );
PLG_linetype( 0, 1.0, 1.0 ); 
PLG_mov( x1, y1 ); PLG_lin( x1, y2 ); PLG_lin( x2, y2 );
PLG_linetype( 0, 0.5, 1.0 );
setptsize( ptextsize );
PLG_mov( x1+nmarg, y1+nmarg );
PLG_dotext( pick[ ipk ].objval, 'L' );
PLG_flush();


return( 0 );
}


/* =============================== */
/* GETGUI_TEXT - render some text.  */
/* Returns obj handle of text zone. */
/* =============================== */
getgui_text( x, y, text, align, valign, size, color, hruling )
double x, y;	/* location of upper left of textchunk */
char *text;	/* the text */
char *align;	/* "center", "left", or "right" */
char *valign;   /* "top", "center", "bottom" */
int size;	/* point size of text */
char *color;	/* color of text */
int hruling; 	/* 0 = none   'h' = horizontal rule at bottom   */
{
int nrows, ncols;
double theight;
int zoneid;


if( relalign == 2 || relalign == 3) x = x_override;
if( relalign == 1 || relalign == 3) y = y_override;
relalign = 0;

setptsize( size );
y -= PLG.curtextheight;
if( valign[0] != 't' ) {
	measuretext( text, &nrows, &ncols );
	theight = nrows * PLG.curtextheight;
	if( valign[0] == 'b' ) y += theight;
	else if( valign[0] == 'c' ) y += theight/2.0;
	}
PLG_color( color );
PLG_init_bb2();   /* initialize an object bounding box */
PLG_mov( x, y );
PLG_dotext( text, toupper( align[0] ));

zoneid = savezone(); /* add an entry to picklist, for sake of relalign.. */

if( hruling ) {
	PLG_color( "white" );
	PLG_mov( pick[ zoneid-1 ].x1, pick[ zoneid-1 ].y1-0.05 ); PLG_lin( pick[ zoneid-1 ].x2, pick[ zoneid-1 ].y1-0.05 );
	PLG_color( "black" );
	PLG_mov( pick[ zoneid-1 ].x1, pick[ zoneid-1 ].y1-0.07 ); PLG_lin( pick[ zoneid-1 ].x2, pick[ zoneid-1 ].y1-0.07 );
	}

return( zoneid );
}

/* ================================= */
/* MEASURETEXT - find #rows and #cols in a chnk of text  (new & improved) */
/* ================================= */
static int
measuretext( txt, nrows, ncols )
char *txt;
int *nrows, *ncols;
{
int i, j;
*nrows = 1;
*ncols = 0;
for( i = 0, j = 0; txt[i] != '\0'; i++, j++ ) {
	if( txt[i] == '\n' && txt[i+1] != '\0' ) {
		(*nrows)++;
		if( j > *ncols ) *ncols = j;
		j = 0;
		}
	}
if( j > *ncols ) *ncols = j;
return( 0 );
}

/* ======================================== */
/* GETGUI_RELALIGN_NEXT - 		    */
/* Tell getgui to align the next rendered object relative to an existing object.   */
/* This overrides aspects of the location given w/ the next object's call          */
/* ======================================== */

getgui_relalign_next( vside, vobjid, vsep, hside, hobjid, hsep )
char *vside; /* "top", "bottom", "center", or "".  If not "", overrides Y */
int vobjid;  /* id of existing object (first=1) to align against vertically */
double vsep; /* amount of separation, vertical */
char *hside; /* "left", "right", "center", or "".  If not "", overrides X */
int hobjid;  /* id of existing object (first=1) to align against horizontally */
double hsep; /* amount of separation, horizontal */
{
int ipk;
relalign = 0;
if( vside[0] == 't' || vside[0] == 'b' || vside[0] == 'c' ) {
	if( vobjid < 1 || vobjid > npick ) { fprintf( stderr, badlistlocmsg, "relalign (v)" ); return( 1 ); }
	relalign += 1;
	ipk = vobjid-1;
	if( pick[ipk].objtype == 'B' ) ipk = pick[ ipk ].iprop3;  /* jump to bb for entire button set */
	if( pick[ipk].objtype == 'N' ) ipk = pick[ ipk ].iprop3 +1; /* jump to bb for entire ns */
	if( vside[0] == 't' ) y_override = pick[ ipk ].y2 + vsep; 
	else if( vside[0] == 'b' ) y_override = pick[ ipk ].y1 - vsep; 
	else if( vside[0] == 'c' ) y_override = pick[ ipk ].y1 + ((pick[ipk].y2-pick[ipk].y1)/2.0);
	}
if( hside[0] == 'l' || hside[0] == 'r' || hside[0] == 'c' ) {
	if( hobjid < 1 || hobjid > npick ) { fprintf( stderr, badlistlocmsg, "relalign (h)" ); return( 1 ); }
	relalign += 2;
	ipk = hobjid-1;
	if( pick[ipk].objtype == 'B' ) ipk = pick[ ipk ].iprop3;  /* jump to bb for entire button set */
	if( pick[ipk].objtype == 'N' ) ipk = pick[ ipk ].iprop3 +1; /* jump to bb for entire ns */
	if( hside[0] == 'l' ) x_override = pick[ ipk ].x1 - hsep; 
	else if( hside[0] == 'r' ) x_override = pick[ ipk ].x2 + hsep; 
	else if( hside[0] == 'c' ) x_override = pick[ ipk ].x1 + ((pick[ipk].x2-pick[ipk].x1)/2.0);
	}

return( 0 );
}

/* ======================================== */
/* GETGUI_SAVEZONE - save a non-interactive zone, so that relaligns can be done against it. */
/* if you want to use relalign against a button set or nameselect box, */
/* this must be called to get the overall zone relalign (since zones are only kept for */
/* individual buttons and selected entries).  Returns obj id of newly created zone */
/* ======================================== */
static int
savezone()
{
if( npick >= MAXPICK-1 ) { fprintf( stderr, fullobjmsg ); return( 0 ); }
/* get obj bounding box */
PLG_get_bb2( &(pick[npick].x1), &(pick[npick].y1), &(pick[npick].x2), &(pick[npick].y2) );
pick[ npick ].objtype = MISC; 
strcpy( pick[ npick ].objval, "" );
npick++;
return( npick );
}

/* ================================= */
/* GETGUI_GETKEY - get a key press or mouse click asyncronously */
/* Adapted from PLG_getkey().. with extra embedded functionality such as cursor blink. */
/* ================================= */
getgui_getkey( x, y, e )
double *x, *y;
int *e;
{
int icount;
PLG.event = 0;

/* note: the following loop has to run quick enough so as not to
   miss any keystrokes of a fast typist.. */
icount = 0;
while( 1 ) {
        Easync();
        if( PLG.event != 0 && PLG.event < 1005 ) {  /* interested in all keys and mouse buttons
                                                but no expose, resize, etc. */
                *x = PLG.eventx; *y = PLG.eventy; *e = PLG.event;
                break;
                }

        usleep( 20000 ); /* loop delay - resulting in 50 cycles per second */
	icount++;

	if( clrfile != NULL && icount % 10 == 9 ) { /* check the clrfile if any - 5 times per second */
		FILE *fp;
		fp = fopen( clrfile, "r" );
                if( fp != NULL ) {
                	fclose( fp );
                       	unlink( clrfile );
			*e = E_MESSAGE;  /* clrfile found */
			return( 0 );
                        }
		}

	if( async && icount % 16 == 15 ) { printf( "!waiting!\n" ); fflush( stdout ); } /* output approx every 1/3 second */

	if( curtextfield && icount % 10 == 9 ) { /* cursor blink every 1/5 of a second */
		cursor_blink();
		}

	if( icount % 50 == 49 ) icount = 0;
        }

return( 0 );
}


/* ================================ */
/* CURSOR_BLINK - blink on/off cursor in current textfield. */
/* ================================ */
/* Not yet implemented */
cursor_blink()
{
int ipk;
if( curtextfield == 0 ) return( 0 );
else	{
	if( cursorstate ) { PLG_color( tcolor ); cursorstate = 0; }
	else cursorstate = 1;
	PLG_mov( cursorx-0.04, cursory ); PLG_lin( cursorx+0.04, cursory ); PLG_mov( cursorx, cursory );
	PLG_lin( cursorx, cursortopy ); PLG_mov( cursorx-0.04, cursortopy ); PLG_lin( cursorx+0.04, cursortopy );
	if( !cursorstate ) PLG_color( "black" );
	}
return(0);
}

/* ================================ */
/* GETGUI_SET - set a parameter     */
/* ================================ */
getgui_set( parm, ival, dval, sval )
char *parm;
int ival;
double dval;
char *sval;
{
if( strcmp( parm, "winx" )==0 ) win_x = ival;
else if( strcmp( parm, "winy" )==0 ) win_y = ival;
else if( strcmp( parm, "winfixw" )==0 ) winfixw = dval;
else if( strcmp( parm, "winfixh" )==0 ) winfixh = dval;
else if( strcmp( parm, "listsep" )==0 ) strcpy( listsep, sval );
else if( strcmp( parm, "bgcolor" )==0 ) strcpy( bgcolor, sval );
else if( strcmp( parm, "winmarg" )==0 ) { winxmarg = winymarg = dval; }
else if( strcmp( parm, "clrfile" )==0 ) clrfile = sval; 
else if( strcmp( parm, "btnsep" )==0 ) bsep = dval;
else if( strcmp( parm, "btndress" )==0 ) {
	if( strcmp( sval, "outline" )==0 ) bdressup = 0;
	else if( strcmp( sval, "bevel" )==0 ) bdressup = 1;
	else if( strcmp( sval, "drop" )==0 ) bdressup = 2;
	}
else if( strcmp( parm, "selsize" )==0 ) ntextsize = ptextsize = ival;
else if( strcmp( parm, "keysize" )==0 ) ttextsize = ival;
else if( strcmp( parm, "async" )==0 ) async = ival;
else { fprintf( stderr, "getgui_set: unrecognized parameter (%s)\n", parm ); return( 1 ); }
return( 0 );
}
